@@ -15,9 +15,18 @@ var CmdAPI = &base.Command{
 		cmdGetStats,
 		cmdQueryStats,
 		cmdSysStats,
+		cmdBalancerInfo,
+		cmdBalancerOverride,
 		cmdAddInbounds,
 		cmdAddOutbounds,
 		cmdRemoveInbounds,
 		cmdRemoveOutbounds,
+		cmdInboundUser,
+		cmdInboundUserCount,
+		cmdAddRules,
+		cmdRemoveRules,
+		cmdSourceIpBlock,
+		cmdOnlineStats,
+		cmdOnlineStatsIpList,
 	},
 }
